local MAJOR = "StarLibLCDText-1.0" 
local MINOR = 11
assert(LibStub, MAJOR.." requires LibStub") 
local LCDText = LibStub:NewLibrary(MAJOR, MINOR)
if not LCDText then return end
local core = LibStub("StarLibCore-1.0")
assert(core, MAJOR .. " requires StarLibCore-1.0")
local LibError = LibStub("StarLibError-1.0")
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibBuffer = LibStub("StarLibBuffer-1.0", true)
assert(LibBuffer, MAJOR .. " requires StarLibBuffer-1.0")
local LibUtils = LibStub("StarLibUtils-1.0", true)
assert(LibUtils, MAJOR .. " requires StarLibUtils-1.0")

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")

local DEFAULT_LIMIT = 50

if not LCDText.__index then
	LCDText.__index = LCDText
end

-- @name StarLCDText:New
-- @usage StarLCDText:New(visitor, rows, cols, yres, xres, layers)
-- @param visitor A StarLibCore object
-- @param rows The number of rows
-- @param cols The number of cols
-- @param yres The y resolution for characters
-- @param xres The x resolution for characters
-- @param layers The number of layers
-- @param errorLevel Error verbosity level
-- @return A new StarLCDText object

function LCDText:New(visitor, rows, cols, yres, xres, layers, errorLevel, realBlit, data)
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
		obj.LayoutFB = {}
		obj.TransitionFB = {}
	end

	setmetatable(obj, self)
	
	tinsert(objects, obj)
	objectsDict[obj] = obj

	obj.error = LibError:New(MAJOR, errorLevel) 
	obj.visitor = visitor
	obj.LROWS = rows
	obj.LCOLS = cols
	obj.DROWS = rows
	obj.DCOLS = cols
	obj.YRES = yres
	obj.XRES = xres
	obj.LAYERS = layers
	obj.errorLevel = errorLevel
	obj.RealBlit = realBlit
	obj.data = data
	obj.transition_tick = 0

	local n = rows * cols
	
	for i = 0, layers do
		obj.LayoutFB[i] = LibBuffer:New(MAJOR .. ".LayoutFB", n, " ", self.errorLEvel)
		obj.TransitionFB[i] = LibBuffer:New(MAJOR .. ".TransitionFB", n, " ", self.errorLevel)
	end
	obj.DisplayFB = LibBuffer:New(MAJOR .. ".DisplayFB", n, "", self.errorLevel)
	
	for n = 0, rows * cols - 1 do
		assert(obj.LayoutFB[0].buffer[n], format("%d", n))
	end
	return obj	
	
end

-- @name StarLCDText:Del
-- @usage StarLCDText:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LCDText:Del(lcd)
	if not lcd then lcd = self end
	
	if type(lcd) == "table" then
		pool[lcd] = true
		lcd.error:Del()
		for i = 0, self.LAYERS do
			obj.LayoutFB[i]:Del()
			obj.TransitionFB[i]:Del()
		end
		lcd.DisplayFB:Del()
	end
end

function LCDText:Clear()
	for i = 0, self.LROWS * self.LCOLS - 1 do
		for l = 0, self.LAYERS - 1 do
			self.LayoutFB[l]:Clear()
			self.TransitionFB[l]:Clear()
		end
	end
	self:TextBlit(0, 0, self.LROWS, self.LCOLS)
end

function LCDText:TextBlit(row, col, height, width)
	--local lr, lc
	--local dr, dc
	--local p1, p2
	local fb = LibBuffer:New(MAJOR .. "._fb", self.LROWS * self.LCOLS, " ", self.errorLevel)
	
	local r = row
	
	while r < self.LROWS and r < row + height do
		local c = col
		while c < self.LCOLS and c < col + width do
			local n = r * self.LCOLS + c
			local l = self.LAYERS - 1
			while l >= 0 do
				if self.LayoutFB[l].buffer[n] ~= ' ' and self.LayoutFB[l].buffer[n] ~= false then
					assert(self.LayoutFB[l].buffer[n], format("1 %d %d", n, self.LROWS * self.LCOLS))
					fb:Replace(n, self.LayoutFB[l].buffer[n] or " ")
					
				end
				l = l - 1
			end
			c = c + 1
		end
		r = r + 1
	end
	
	--fb:Copy(self.LayoutFB[0])
	for i = 0, fb:Size() - 1 do
		assert(fb.buffer[i], format("2 %d %d", i, fb:Size()))
	end

	--/* loop over layout rows */
	local lr = row
	
	while lr < self.LROWS and lr < row + height do
		--/* transform layout to display row */
		local dr = lr;
		--/* sanity check */
		if dr >= 0 and dr < self.DROWS then 		
			--/* loop over layout cols */
			local lc = col
			while lc < self.LCOLS and lc < col + width do
				--/* transform layout to display column */
				local dc = lc;
				--/* sanity check */
				if dc >= 0 and dc < self.DCOLS then
					--/* find start of difference */
					if self.DisplayFB.buffer[dr * self.DCOLS + dc] ~= fb.buffer[lr * self.LCOLS + lc] then
					--if string.sub(self.DisplayFB, dr * self.DCOLS + dc, 1) ~= string.sub(fb, lr * self.LCOLS + lc) then
						--/* find end of difference */
						local p1 = dc
						local p2 = p1
						while lc < self.LCOLS and lc < col + width do
							--/* transform layout to display column */
							dc = lc;
							--/* sanity check */
							if dc >= 0 and dc < self.DCOLS then
								if self.DisplayFB.buffer[dr * self.DCOLS + dc] ~= fb.buffer[lr * self.LCOLS + lc] then
									p2 = dc;
								end
							end
							lc  = lc + 1
						end

						--/* send to display */
						local from = fb:Sub(lr * self.LCOLS + p1)
						self.DisplayFB:Memcopy(dr * self.DCOLS + p1, from, 0, p2 - p1 + 1)
						--self.DisplayFB:Merge(from, dr * self.DCOLS + p1, p2 - p1 + 1)
						if (self.RealBlit) then
							local n = dr * self.DCOLS + p1
							local tmp = self.DisplayFB:Sub(n, n + p2 - p1)
							self:RealBlit(self.data, dr, p1, tmp.buffer, p2 - p1 + 1)
							tmp:Del()
						end
						from:Del()
					end
				end
				lc = lc + 1
			end
		end
		lr = lr + 1
	end
	fb:Del()
end

function LCDText.TextDraw(widget, data)
	local lcd = widget.visitor.lcd
	local n  = widget.widget.row * lcd.LCOLS + widget.widget.col
	local len = strlen(widget.buffer)
	local num = 1
	local layer = widget.widget.layer
    local fb;
    if lcd.transitioning and widget.widget.layout_base == lcd.transitionLayout then
        fb = lcd.TransitionFB[layer];
    else
        fb = lcd.LayoutFB[layer];
	end
	
	for i = n, n + len do
		if i - widget.widget.row * lcd.LCOLS >= lcd.LCOLS then
			break
		end
		fb:Replace(i, widget.buffer:sub(num, num))
		num = num + 1
	end

	if not lcd.transitioning then
		lcd:TextBlit(widget.widget.row, widget.widget.col, 1, len)
	end
end

function LCDText.BarDraw(widget)
	local lcdText = widget.widget.visitor.lcd
    local length = widget.length
    local row = widget.widget.row
    local col = widget.widget.col
    local layer = widget.widget.layer

    local val = math.floor((widget.val1 * length) + 0.5 );

    if(val < 0) then val = 0 end

    local n = row * lcdText.LCOLS + col;

    local fb;
    if lcdText.transitioning and widget.widget.layout_base == lcdText.transitionLayout then
        fb = lcdText.TransitionFB[layer];
    else
        fb = lcdText.LayoutFB[layer];
	end

    for x = n, n + length do
        fb.buffer[x] = ' ';
	end

    if( widget.direction == widget.DIR_EAST ) then
		for i = n, n + val - 1 do
			if mod(i, lcdText.LCOLS) >= lcdText.LCOLS then
				break
			end
            fb.buffer[i] = "#"
        end
    elseif( widget.direction == widget.DIR_WEST ) then
        for i = n, n - val + 1, -1 do
			if i / lcdText.LROWS < row then
				break
			end
			fb.buffer[i] = "#"
        end
	end
    if not lcdText.transitioning then
        lcdText:TextBlit(row, col, 1, length)
	end
end

function LCDText.HistogramDraw(widget)
	local lcdText = widget.widget.visitor.lcd

    local row = widget.widget.row
    local col = widget.widget.col
    local width = widget.width
	local height = widget.height
	local layer = widget.widget.layer

    local fb;
    if(lcdText.transitioning and widget.widget.layout_base == lcdText.transitionLayout) then
        fb = lcdText.TransitionFB[layer]
    else
        fb = lcdText.LayoutFB[layer]
	end
	
	if not fb then 
		widget.error:Print("Bad layer " .. layer)
		return 
	end

	for c = col, col + width - 1 do
		for r = row, row + height - 1 do
			fb.buffer[r * lcdText.LCOLS + c] = " "
		end
	end
	
	for c = col, col + width - 1 do
		local val
		if c - col >= #widget.history then
			val = 0
		else
			val = floor(widget.history[c - col + 1] * height + 0.5)
		end
		if not widget.reversed then
			for r = row, row + val - 1 do
				fb.buffer[(row + r) * lcdText.LCOLS + c] = widget.char or "#"
			end
		else
		
			for r = row + height, row + height - val, -1 do
				fb.buffer[(row + r) * lcdText.LCOLS + c] = widget.char or "#"
			end
		end
	end
	
    if(not lcdText.transitioning) then
        lcdText:TextBlit(row, col, height, width);
	end
end

function LCDText:SignalTransitionStart(layout)
	self.transitioning = true; 
	self.transitionLayout = layout
end

function LCDText:SignalTransitionEnd()
	self.transitioning = false
end

function LCDText:Transition()
	if self.visitor.direction == self.visitor.TRANSITION_LEFT or
	   self.visitor.direction == self.visitor.TRANSITION_RIGHT or
	   self.visitor.direction == self.visitor.TRANSITION_BOTH then
		self:TransitionLeftRight()
	elseif self.visitor.direction == self.visitor.TRANSITION_UP or
	       self.visitor.direction == self.visitor.TRANSITION_DOWN then
		self:TransitionUpDown()
	else
		self:TransitionLeftRight()
	end
end

function LCDText:TransitionLeftRight()
	local direction = self.visitor.direction
	local col
	local left, right
	local layout = LibBuffer:New("tmplayout", self.LROWS * self.LCOLS, " ", self.errorLevel)
	local transition = LibBuffer:New("tmptransition", self.LROWS * self.LCOLS, " ", self.errorLevel)

	for l = self.LAYERS - 1, 0, -1 do
		for n = 0, self.LROWS * self.LCOLS - 1 do
			if self.LayoutFB[l].buffer[n] ~= ' ' then
				layout.buffer[n] = self.LayoutFB[l].buffer[n]
			end
			if self.TransitionFB[l].buffer[n] ~= ' ' then
				transition.buffer[n] = self.TransitionFB[l].buffer[n]
			end
		end
	end

    for row = 0, self.LROWS - 1 do
		local n = row * self.LCOLS
		
		
        if( direction == self.visitor.TRANSITION_LEFT or
            (direction == self.visitor.TRANSITION_BOTH and mod(row, 2) == 0)) then
            col = self.LCOLS - self.transition_tick - 1;
            left = layout;
            right = transition;
        elseif( direction == self.visitor.TRANSITION_RIGHT or direction == self.visitor.TRANSITION_BOTH) then
            col = self.transition_tick;
            left = transition;
            right = layout;
        else 
            col = self.transition_tick;
            left = transition;
            right = layout;
            self.error:Print("LCDText: Bad transition direction: " .. direction);
        end
		--self.DisplayFB:Memset(n, ' ', self.LCOLS)
		self.DisplayFB:Memcopy(n, left, n + (self.LCOLS - col), col)
		self.DisplayFB:Memcopy(n + col + 1, right, n, self.LCOLS - col - 1)
        --memset(DisplayFB + n, ' ', LCOLS);
        --memcpy(DisplayFB + n, left + n + (LCOLS - col), col);
        --memcpy(DisplayFB + n + col + 1, right + n, LCOLS - col - 1);
        if(direction ~= self.visitor.TRANSITION_BOTH) then
            self.DisplayFB.buffer[n + col] = ':';
		end
        if(self.RealBlit) then
			local fb = self.DisplayFB:Sub(n)
            self:RealBlit(self.data, row, 0, fb.buffer, self.LCOLS);
			fb:Del()
		end
		
    end

	self.transition_tick = self.transition_tick + 1
    if( self.transition_tick >= self.LCOLS ) then
        self.transitioning = false;
        self.transition_tick = 0;
        --emit static_cast<LCDEvents *>(
        --    visitor_->GetWrapper())->_TransitionFinished();
		self.visitor:TransitionFinished()
        for l = 0, self.LAYERS - 1 do
			self.LayoutFB[l]:Copy(self.TransitionFB[l])
			--self.LayoutFB[l]:Memcopy(0, self.TransitionFB[l], 0, self.LCOLS * self.LROWS)
            --memcpy(LayoutFB[l], TransitionFB[l], LCOLS * LROWS);
			self.TransitionFB[l]:Clear()
			--self.TransitionFB[l]:Memset(0, ' ', self.LCOLS * self.LROWS)
            --memset(TransitionFB[l], ' ', LCOLS * LROWS);
        end
        self:TextBlit(0, 0, self.LROWS, self.LCOLS);
    end
	
	layout:Del()
	transition:Del()
end

function LCDText:TransitionUpDown()
	do return end
    local direction = self.visitor.direction
    local row;
    local top, bottom;
    local layout = LibBuffer:New("tmplayout", LROWS * LCOLS, " ", self.errorLevel)
	local transition = LibBuffer:New("tmptransition", LROWS * LCOLS, " ", self.errorLevel)
 
    for l = self.LAYERS - 1, 0, -1 do
        for n = 0, self.LROWS * self.LCOLS - 1 do
            if(self.LayoutFB[l].buffer[n] ~= ' ') then
                layout.buffer[n] = self.LayoutFB[l].buffer[n];
			end
            if(self.TransitionFB[l].buffer[n] ~= ' ') then
                transition.buffer[n] = self.TransitionFB[l].buffer[n];
			end
        end
    end

    if(direction == self.visitor.TRANSITION_UP) then
        top = layout;
        bottom = transition;
        row = self.LROWS - self.transition_tick
    else 
        top = transition;
        bottom = layout;
        row = self.transition_tick
    end

    for r = 0, self.LROWS - 1 do
        local n = r * LCOLS;
        for l = 0, self.LAYERS do
			self.DisplayFB:Memset(n, ' ', self.LCOLS)
            if(r < row) then
				self.DisplayFB:Memcopy(n, top, n, self.LCOLS)
            else
				self.DisplayFB:Memcopy(n, bottom, n, self.LCOLS)
            end
        end
        if(self.RealBlit) then
            RealBlit(self.visitor.lcd, r, 0, self.DisplayFB + n, self.LCOLS);
		end
    end

	self.transition_tick = self.transition_tick + 1
    if( self.transition_tick >= self.LROWS ) then
        self.transitioning = false;
        self.transition_tick = 0;
        --emit static_cast<LCDEvents *>(
        --    visitor_->GetWrapper())->_TransitionFinished();
		self.visitor:TransitionFinished()
        for l = 0, self.LAYERS - 1 do
			self.LayoutFB[l]:Memcopy(0, self.TransitionFB[l], 0, self.LCOLS * self.LROWS)
            --memcpy(LayoutFB[l], TransitionFB[l], LCOLS * LROWS);
			self.TransitionFB[l]:Memset(0, ' ', self.LCOLS * self.LROWS)
            --memset(TransitionFB[l], ' ', LCOLS * LROWS);
        end
        self:TextBlit(0, 0, self.LROWS, self.LCOLS);
    end
	
	layout:Del()
	transition:Del()
end
